////////////////////////////////////////////////////////////////////////
//
//      Michelle Moravan (moravan@cs.wisc.edu)
//      CS department U of Wisconsin
//
//      CS 701
//      October 2005
//
//      reg_alloc.cc
//
////////////////////////////////////////////////////////////////////////
//
// Contains all functions to be used for register allocation.
//
////////////////////////////////////////////////////////////////////////

//----- Include -----
#include <assert.h>
#include <iostream.h>
#include <stdio.h>

extern "C" {
#include <simple.h>
}

#include "reg_alloc.h"

void print_instr(simple_instr* instr) {
    cout << "*** addr=0x" << instr << " type=";
    switch (simple_op_format(instr->opcode)) {
    case 0:
      cout << "BASE";
      break;
    case 1:
      cout << "BJ " << instr->u.bj.target->name;
      break;
    case 2:
      cout << "LDC";
      break;
    case 3:
      cout << "CALL";
      break;
    case 4:
      cout << "MBR";
      break;
    case 5:
      cout << "LABEL " << instr->u.label.lab->name;
      break;
    }
    cout << endl;
}

void debug_instr_types(simple_instr* inlist) {
  simple_instr* cur_instr = inlist; // current instruction
  while (cur_instr != NULL) {
    print_instr(cur_instr);
    cur_instr = cur_instr->next;
  }
}

basic_block* create_block() {
  basic_block* block = new basic_block;
  block->pred_head = NULL;
  block->pred_tail = NULL;
  block->succ_head = NULL;
  block->succ_tail = NULL;
  block->instr_head = NULL;
  block->instr_tail = NULL;
  return block;
}

basic_block_node* create_node() {
  basic_block_node* node = new basic_block_node;
  node->block = NULL;
  node->next = NULL;
  return node;
}

void add_instr(basic_block* block, simple_instr* instr) {
  if (block->instr_head == NULL) {
    block->instr_tail = block->instr_head = instr;
  } else {
    assert(block->instr_tail->next == instr);
    block->instr_tail = block->instr_tail->next;
  }
}

void add_block_pred(basic_block* cur, basic_block* pred) {
  if (cur->pred_head == NULL) {
    cur->pred_head = create_node();
    cur->pred_head->block = pred;
    cur->pred_tail = cur->pred_head;
  } else {
    cur->pred_tail->next = create_node();
    cur->pred_tail = cur->pred_tail->next;
    cur->pred_tail->block = pred;
  }
}

void add_block_succ(basic_block* cur, basic_block* succ) {
  if (cur->succ_head == NULL) {
    cur->succ_head = create_node();
    cur->succ_head->block = succ;
    cur->succ_tail = cur->succ_head;
  } else {
    cur->succ_tail->next = create_node();
    cur->succ_tail = cur->succ_tail->next;
    cur->succ_tail->block = succ;
  }
}

basic_block_node* build_cfg(simple_instr* inlist) {
  simple_instr* cur_instr = inlist; // current instruction
  int instr_format = -1;
  bool need_new_block = true;
  bool unconditional = false;
  basic_block* cur_block = NULL; // new block, before it's linked into
                                 // the CFG

  // A linked list of all blocks in the CFG
  // The CFG itself is embedded in the pred/succ components of the blocks
  basic_block_node* cfg_head = NULL;
  basic_block_node* cfg_tail = NULL;

  // Phase 1: Determine all basic blocks and create arcs between those
  // linked sequentially
  while (cur_instr != NULL) {
    instr_format = simple_op_format(cur_instr->opcode);
    if (((instr_format == BASE_FORM || instr_format == LDC_FORM || 
          instr_format == CALL_FORM) && need_new_block) || 
        instr_format == LABEL_FORM) {
      cur_block = create_block();
      if (cfg_head == NULL) {
        cfg_head = create_node();
        cfg_head->block = cur_block;
        cfg_head->next = NULL;
        cfg_tail = cfg_head;
      } else {
        if (!unconditional) {
          add_block_succ(cfg_tail->block, cur_block);
          add_block_pred(cur_block, cfg_tail->block);
        }
        cfg_tail->next = create_node();
        cfg_tail = cfg_tail->next;
        cfg_tail->block = cur_block;
      }
      need_new_block = false;
      unconditional = false;
    } else if (instr_format == BJ_FORM || instr_format == MBR_FORM) { 
      need_new_block = true;
      if (strcmp(simple_op_name(cur_instr->opcode), "jmp ") == 0) {
        unconditional = true;
      }
    }
    add_instr(cur_block, cur_instr);
    cur_instr = cur_instr->next;
  }
  
  // Phase 2
  basic_block_node* cfg_cur = cfg_head;
  basic_block* target_block = NULL;
  while (cfg_cur != NULL) {
    cur_instr = cfg_cur->block->instr_tail;
    instr_format = simple_op_format(cur_instr->opcode);
    if (instr_format == BJ_FORM) {
      target_block = find_target_block(cfg_head, cur_instr->u.bj.target);
      assert(target_block != NULL);
      add_block_succ(cfg_cur->block, target_block);
      add_block_pred(target_block, cfg_cur->block);
    } else if (instr_format == MBR_FORM) {
      for (unsigned i = 0; i < cur_instr->u.mbr.ntargets; i++) {
        target_block = find_target_block(cfg_head, 
                                         cur_instr->u.mbr.targets[i]);
        assert(target_block != NULL);
        add_block_succ(cfg_cur->block, target_block);
        add_block_pred(target_block, cfg_cur->block);
      }
    } 
    cfg_cur = cfg_cur->next;
  }
  return cfg_head;
}

basic_block* find_target_block(basic_block_node* block_list, 
                               simple_sym* target) {
  assert(target->kind == LABEL_SYM);
  basic_block_node* cur = block_list;
  while (cur != NULL) {
    if (simple_op_format(cur->block->instr_head->opcode) == LABEL_FORM && 
      cur->block->instr_head->u.label.lab->name == target->name) {
      break;
    }
    cur = cur->next;
  }
  return cur->block;
}

void print_cfg(basic_block_node* cfg) {
  basic_block_node* cur_node = cfg;
  basic_block_node* succ_node = NULL;
  basic_block_node* pred_node = NULL;
  simple_instr* cur_instr = NULL;
  cout << "***** CFG *****" << endl;

  while (cur_node != NULL) {
    cout << "*** BASIC BLOCK " << cur_node->block << " ***" << endl;

    // print contents
    cout << "\t*** INSTRS ***" << endl;
    cur_instr = cur_node->block->instr_head;
    while (cur_instr != cur_node->block->instr_tail) {
      cout << "\t";
      print_instr(cur_instr);
      cur_instr = cur_instr->next;
    }
    if (cur_node->block->instr_head != cur_node->block->instr_tail) {
      cout << "\t";
      print_instr(cur_instr); // because tail is the last valid instr
    }

    // print successors
    cout << "\t*** SUCCS ***" << endl;
    succ_node = cur_node->block->succ_head;
    while (succ_node != NULL) {
      cout << "\t" << succ_node->block << endl;
      succ_node = succ_node->next;
    }

    // print predecessors
    cout << "\t*** PREDS ***" << endl;
    pred_node = cur_node->block->pred_head;
    while (pred_node != NULL) {
      cout << "\t" << pred_node->block << endl;
      pred_node = pred_node->next;
    }

    cur_node = cur_node->next;
  }
}

void delete_cfg(basic_block_node* cfg) {
  basic_block_node* prev = cfg;
  basic_block_node* cur = prev->next;
  while (cur != NULL) {
    delete_block(prev->block);
    delete prev;
    prev = cur;
    cur = cur->next;
  }
}

// doesn't deallocate the blocks pointed to by the predecessors &
// successors, since multiple links to these exist
// the only way to ensure each is deallocated exactly once is to use
// the references in the cfg itself
// similarly, instructions are not deallocated for the same reason;
// their unique references are in inlist
void delete_block(basic_block* block) {
  basic_block_node* prev;
  basic_block_node* cur;

  // deallocate the predecessors
  prev = block->pred_head;
  while (prev != NULL) {
    cur = prev->next;
    delete prev;
    prev = cur;
  }

  // deallocate the successors
  prev = block->succ_head;
  while (prev != NULL) {
    cur = prev->next;
    delete prev;
    prev = cur;
  }
}
