/************************************************************************
 * hash_map.h: implementation of hashed maps.
************************************************************************/





#if !defined(_hash_map_h_)
#define _hash_map_h_





/************************************************************************
 * header files.
************************************************************************/

extern "C" {
#include <assert.h>
}

#include "hash_table.h"





/************************************************************************
 * template<class Key, class Value, class Hash> class hash_map
************************************************************************/

template<class Key, class Value, class Hash>
class hash_map {
private:
    typedef hash_map<Key,Value,Hash>   self;
    typedef local_hash_table<Key,Value,Hash> rep_type;
    typedef rep_type::iterator         rep_iterator;
    typedef rep_type::const_iterator   rep_const_iterator;


public:
    typedef rep_type::key_type           key_type;
    typedef rep_type::key_hash           key_hash;
    typedef Value                        data_type;
    typedef       pair<CONST Key,Value>  value_type;
    typedef       pair<CONST Key,Value>& reference;
    typedef const pair<CONST Key,Value>& const_reference;
    typedef size_t                       size_type;


    class iterator {
    private:
        rep_iterator ri_;

    public:
        explicit iterator() :
            ri_() {
        }

        iterator(const rep_iterator& ri) :
            ri_(ri) {
        }

        iterator(const iterator& i) :
            ri_(i.ri_) {
        }

        const iterator& operator=(const iterator& i) {
            if (this != &i) {
                ri_ = i.ri_;
            }
            return *this;
        }

        ~iterator() {
        }


        reference operator*() {
            return *ri_;
        }

        iterator& operator++() {
            ++ri_;
            return *this;
        }

        iterator operator++(int) {
            iterator tmp = *this;
            operator++();
            return tmp;
        }

        bool operator==(const iterator& i) const {
            return ((this == &i) || (ri_ == i.ri_));
        }
    };

    typedef iterator reverse_iterator;


    class const_iterator {
    private:
        rep_const_iterator ri_;

    public:
        explicit const_iterator() :
            ri_() {
        }

        const_iterator(const rep_const_iterator& ri) :
            ri_(ri) {
        }

        const_iterator(const const_iterator& i) :
            ri_(i.ri_) {
        }

        const const_iterator& operator=(const const_iterator& i) {
            if (this != &i) {
                ri_ = i.ri_;
            }
            return *this;
        }

        ~const_iterator() {
        }


        const_reference operator*() const {
            return *ri_;
        }

        const_iterator& operator++() {
            ++ri_;
            return *this;
        }

        const_iterator operator++(int) {
            const_iterator tmp = *this;
            operator++();
            return tmp;
        }

        bool operator==(const const_iterator& i) const {
            return ((this == &i) || (ri_ == i.ri_));
        }
    };

    typedef const_iterator const_reverse_iterator;


private:
    rep_type rep_;


public:
    iterator begin() {
        return rep_.begin();
    }

    const_iterator begin() const {
        return rep_.begin();
    }

    iterator end() {
        return rep_.end();
    }

    const_iterator end() const {
        return rep_.end();
    }

    reverse_iterator rbegin() {
        return begin();
    }

    const_reverse_iterator rbegin() const {
        return begin();
    }

    reverse_iterator rend() {
        return end();
    }

    const_reverse_iterator rend() const {
        return end();
    }


    size_type size() const {
        return rep_.size();
    }

    size_type count(const key_type& key) const {
        return rep_.count(key);
    }

    bool empty() const {
        return (size() == 0);
    }

    void swap(self& h) {
        ::swap(rep_, h.rep_);
    }


    iterator insert(const value_type& kv) {
        return rep_.insert(kv);
    }

    iterator insert(const_iterator, const value_type& kv) {
        return insert(kv);
    }

    void insert(const_iterator curr, const_iterator last) {
        for (; curr != last; ++curr) {
            insert(*curr);
        }
    }

    pair<iterator,bool> uniq_insert(const value_type& kv) {
        pair<rep_type::iterator,bool> ret = rep_.uniq_insert(kv);
        return make_pair(iterator(ret.first), ret.second);
    }


    size_type erase(const key_type& key) {
        return rep_.erase(key);
    }

    void erase(iterator curr) {
        erase((*curr).first);
    }

    void erase(iterator curr, iterator last) {
        for (; curr != last; ++curr) {
            erase(curr);
        }
    }


    iterator find(const key_type& key) {
        return rep_.find(key);
    }

    const_iterator find(const key_type& key) const {
        return rep_.find(key);
    }


    data_type& operator[](const key_type& key) {
        return rep_[key];
    }

    const data_type& operator[](const key_type& key) const {
        return rep_[key];
    }


    key_hash hashf() const {
        return rep_.hashf();
    }


    bool operator==(const self& h) const {
        return ((this == &h) || (rep_ == h.rep_));
    }

    bool operator!=(const self& h) const {
        return ((this != &h) && (rep_ != h.rep_));
    }


    hash_map(key_hash hf) :
        rep_(hf) {
    }

    hash_map(const_iterator from, const_iterator last,
        key_hash hf) :
        rep_(hf) {
        for (; from != last; ++from) {
            insert(*from);
        }
    }

    hash_map(const self& h) :
        rep_(h.rep_) {
    }

    const self& operator=(const self& h) {
        if (this != &h) {
            rep_ = h.rep_;
        }
        return *this;
    }

    ~hash_map() {
    }
};





#endif /* !defined(_hash_map_h_) */
