/************************************************************************
 * hash_table.h: implementation of linear-hash tables.
************************************************************************/





#if !defined(_hash_table_h_)
#define _hash_table_h_





/************************************************************************
 * header files.
************************************************************************/

extern "C" {
#include <assert.h>
}

#include <cstddef>
#include <stl.h>

#include "hash_utils.h"





/************************************************************************
 * template<class Key, class Data, class Hash> class local_hash_table
************************************************************************/

#define CONST

template<class Key, class Data, class Hash>
class local_hash_table {
private:
    typedef local_hash_table<Key,Data,Hash> self;

    typedef pair<Key,Data>           kd_pair;
    typedef pair<CONST Key,Data>     ckd_pair;
    typedef pair<hash_t,kd_pair>     hkd_pair;
    typedef vector<hkd_pair>         hkd_chain;
    typedef vector<hkd_chain>        hkd_data;

    typedef pair<size_t,size_t>      hash_position;
    typedef pair<bool,hash_position> key_location;

    static const size_t DEFAULT_NBUCKETS   = 32;
    static const size_t DEFAULT_CHAIN_SIZE = 32;


public:
    typedef Key                         key_type;
    typedef Hash                        key_hash;
    typedef Data                        data_type;
    typedef       pair<CONST Key,Data>  value_type;
    typedef       pair<CONST Key,Data>& reference;
    typedef const pair<CONST Key,Data>& const_reference;
    typedef size_t                      size_type;


private:
    key_hash       hashf_;
    size_t         chain_size_;
    hkd_data       data_;
    vector<size_t> llevel_;
    size_t         nbuckets_;
    size_t         glevel_;
    size_t         next_;


public:
    class iterator {
    private:
        hkd_data* hkd_data_;
        size_t    chain_;
        size_t    curr_;

    private:
        size_t next_non_empty(size_t from) const {
            assert(hkd_data_ != 0);
            for (size_t ch = from; ch < hkd_data_->size(); ++ch) {
                if ((*hkd_data_)[ch].size() != 0) {
                return ch;
                }
            }
            return hkd_data_->size();
        }

    public:
        explicit iterator() :
            hkd_data_(0), chain_(0), curr_(0) {
        }

        iterator(hkd_data* hkdd, size_t chain =0, size_t curr =0) :
            hkd_data_(hkdd), chain_(chain), curr_(curr) {
            assert(hkd_data_ != 0);
            chain_ = next_non_empty(chain_);
        }

        iterator(const iterator& i) :
            hkd_data_(i.hkd_data_), chain_(i.chain_), curr_(i.curr_) {
            chain_ = next_non_empty(chain_);
        }

        ~iterator() {
        }

        ckd_pair& operator*() {
            assert(hkd_data_ != 0);
            return (*hkd_data_)[chain_][curr_].second;
        }

        iterator& operator++() {
            assert(hkd_data_ != 0);
            assert(chain_ <= hkd_data_->size());
            if (chain_ < hkd_data_->size()) {
                assert(curr_ < (*hkd_data_)[chain_].size());
                ++curr_;
                if (curr_ == (*hkd_data_)[chain_].size()) {
                chain_ = next_non_empty(1+chain_);
                curr_  = 0;
                }
            }
            return *this;
        }

        iterator operator++(int) {
            iterator tmp = *this;
            operator++();
            return tmp;
        }

        bool operator==(const iterator& i) const {
            return ((this == &i)
                || ((hkd_data_ == i.hkd_data_)
                && (chain_ == i.chain_) && (curr_ == i.curr_)));
        }
    };

    typedef iterator reverse_iterator;


    class const_iterator {
    private:
        const hkd_data* hkd_data_;
        size_t          chain_;
        size_t          curr_;

    private:
        size_t next_non_empty(size_t from) const {
            assert(hkd_data_ != 0);
            for (size_t ch = from; ch < hkd_data_->size(); ++ch) {
                if ((*hkd_data_)[ch].size() != 0) {
                    return ch;
                }
            }
            return hkd_data_->size();
        }

    public:
        explicit const_iterator() :
            hkd_data_(0), chain_(0), curr_(0) {
        }

        const_iterator(const hkd_data* hkdd, size_t chain, size_t curr) :
            hkd_data_(hkdd), chain_(chain), curr_(curr) {
            assert(hkd_data_ != 0);
            chain_ = next_non_empty(chain_);
        }

        const_iterator(const const_iterator& i) :
            hkd_data_(i.hkd_data_), chain_(i.chain_), curr_(i.curr_) {
            chain_ = next_non_empty(chain_);
        }

        ~const_iterator() {
        }

        const kd_pair& operator*() const {
            assert(hkd_data_ != 0);
            return (*hkd_data_)[chain_][curr_].second;
        }

        const_iterator& operator++() {
            assert(hkd_data_ != 0);
            assert(chain_ <= hkd_data_->size());
            if (chain_ < hkd_data_->size()) {
                assert(curr_ < (*hkd_data_)[chain_].size());
                ++curr_;
                if (curr_ == (*hkd_data_)[chain_].size()) {
                    chain_ = next_non_empty(1+chain_);
                    curr_  = 0;
                }
            }
            return *this;
        }

        const_iterator operator++(int) {
            const_iterator tmp = *this;
            operator++();
            return tmp;
        }

        bool operator==(const const_iterator& i) const {
            return ((this == &i)
                || ((hkd_data_ == i.hkd_data_)
                    && (chain_ == i.chain_) && (curr_ == i.curr_)));
        }
    };
    typedef const_iterator const_reverse_iterator;


private:
    void invariants() const {
#if defined(CHECK_INVARIANTS)
        assert(data_.size() == llevel_.size());

        assert(nbuckets_ <= data_.size());
        assert(data_.size() < (2*nbuckets_));
        assert(next_ < nbuckets_);
        assert(data_.size() == (nbuckets_ + next_));

        for (size_t ch = 0; ch < next_; ++ch) {
            assert(llevel_[ch] == (glevel_+1));
            assert(llevel_[ch] == llevel_[ch+nbuckets_]);
        }
        for (size_t ch = next_; ch < nbuckets_; ++ch) {
            assert(llevel_[ch] == glevel_);
        }
#endif /* defined(CHECK_INVARIANTS) */
    }

    size_t chain_of(hash_t hash) const {
        size_t chain = hash % nbuckets_;
        if (llevel_[chain] > glevel_) {
            chain = hash % (nbuckets_*2);
            assert(chain < data_.size());
        }
        return chain;
    }

    key_location locate(const key_type& key) const {
        hash_t hash  = hashf_(key);
        size_t chain = chain_of(hash);
        for (size_t i = 0; i < data_[chain].size(); ++i) {
            if ((hash == data_[chain][i].first)
                && (key == data_[chain][i].second.first)) {
                return make_pair(true, make_pair(chain, i));
            }
        }
        return make_pair(false, make_pair(size_t(chain), size_t(0)));
    }

    data_type& insert_key(const key_type& key, hash_position hp) {
        hash_t hash  = hashf_(key);
        size_t chain = hp.first;

        data_[chain].push_back(hkd_pair(hash, make_pair(key, data_type())));
        if (data_[chain].size() > chain_size_) {
            split();
        }

        key_location loc = locate(key);
        assert(loc.first);
        return data_[loc.second.first][loc.second.second].second.second;
    }

    void remove(size_t chain, size_t i) {
        data_[chain].erase(&data_[chain][i]);
    }

    void split() {
        size_t nch = data_.size();
        data_.push_back(hkd_chain());
        llevel_.push_back(++llevel_[next_]);

        hkd_chain& ch = data_[next_];
        hkd_chain  tmp;
        for (size_t i = 0; i < ch.size(); ++i) {
            size_t nchain = chain_of(ch[i].first);
            assert((nchain == next_) || (nchain == nch));
            if (nchain == next_) {
                tmp.push_back(ch[i]);
            }
            else {
                data_[nch].push_back(ch[i]);
            }
        }
        assert((tmp.size()+data_[nch].size()) == ch.size());
        ch = tmp;

        ++next_;
        if (next_ == nbuckets_) {
            next_ = 0;
            nbuckets_ *= 2;
            ++glevel_;
        }
    }


public:
    iterator begin() {
        return iterator(&data_, size_t(0), size_t(0));
    }

    const_iterator begin() const {
        return const_iterator(&data_, size_t(0), size_t(0));
    }

    iterator end() {
        return iterator(&data_, data_.size(), size_t(0));
    }

    const_iterator end() const {
        return const_iterator(&data_, data_.size(), size_t(0));
    }

    reverse_iterator rbegin() {
        return begin();
    }

    const_reverse_iterator rbegin() const {
        return begin();
    }

    reverse_iterator rend() {
        return end();
    }

    const_reverse_iterator rend() const {
        return end();
    }


    size_type size() const {
        size_type sz = 0;
        for (size_t chain = 0; chain < data_.size(); ++chain) {
            sz += data_[chain].size();
        }
        return sz;
    }

    size_type count(const key_type& key) const {
        return ((locate(key).first) ? 1 : 0);
    }

    bool empty() const {
        return (size() == 0);
    }

    void swap(self& h) {
        ::swap(hashf_     , h.hashf_);
        ::swap(chain_size_, h.chain_size_);
        ::swap(data_      , h.data_);
        ::swap(llevel_    , h.llevel_);
        ::swap(nbuckets_  , h.nbuckets_);
        ::swap(glevel_    , h.glevel_);
        ::swap(next_      , h.next_);
    }


    iterator insert(const value_type& kd) {
        key_location loc = locate(kd.first);
        if (loc.first) {
            data_[loc.second.first][loc.second.second].second.second = kd.second;
            return iterator(&data_, loc.second.first, loc.second.second);
        }
        insert_key(kd.first, loc.second) = kd.second;
        loc = locate(kd.first);
        return iterator(&data_, loc.second.first, loc.second.second);
    }

    iterator insert(const_iterator, const value_type& kd) {
        return insert(kd);
    }

    void insert(const_iterator curr, const_iterator last) {
        for (; curr != last; ++curr) {
            insert(*curr);
        }
    }

    pair<iterator,bool> uniq_insert(const value_type& kd) {
        key_location loc = locate(kd.first);
        return ((loc.first)
            ? (make_pair(iterator(&data_, loc.second.first, loc.second.second),
                false))
            : (make_pair(insert(kd), true)));
    }


    size_type erase (const key_type& key) {
        key_location loc = locate(key);
        if (loc.first) {
            remove(loc.second.first, loc.second.second);
            return 1;
        }
        return 0;
    }

    void erase(iterator curr) {
        erase((*curr).first);
    }

    void erase(iterator curr, iterator last) {
        vector<key_type> tmp;
        for (; curr != last; ++curr) {
            tmp.push_back((*curr).first);
        }
        for (vector<key_type>::const_iterator curr = tmp.begin();
            curr != tmp.end(); ++curr) {
            erase(*curr);
        }
    }


    iterator find(const key_type& key) {
        key_location loc = locate(key);
        return ((loc.first)
            ? (iterator(&data_, loc.second.first, loc.second.second))
            : (end()));
    }

    const_iterator find(const key_type& key) const {
        key_location loc = locate(key);
        return ((loc.first)
            ? (const_iterator(&data_, loc.second.first, loc.second.second))
            : (end()));
    }


    data_type& operator[](const key_type& key) {
        key_location loc = locate(key);
        if (loc.first) {
            size_t ch = loc.second.first;
            size_t i  = loc.second.second;
            return data_[ch][i].second.second;
        }
        return insert_key(key, loc.second);
    }

    const data_type& operator[](const key_type& key) const {
        key_location loc = locate(key);
        assert(loc.first);
        size_t ch = loc.second.first;
        size_t i  = loc.second.second;
        return data_[ch][i].second.second;
    }


    key_hash hashf() const {
        return hashf_;
    }


    bool operator==(const self& h) const {
        if (size() != h.size()) {
            return false;
        }
        for (const_iterator curr = h.begin(); curr != h.end(); ++curr) {
            if (!locate((*curr).first).first) {
                return false;
            }
        }
        for (const_iterator curr = begin(); curr != end(); ++curr) {
            if (!h.locate((*curr).first).first) {
                return false;
            }
        }
        return true;
    }

    bool operator!=(const self& h) const {
        if (size() != h.size()) {
            return true;
        }
        for (const_iterator curr = h.begin(); curr != h.end(); ++curr) {
            if (!locate((*curr).first).first) {
                return true;
            }
        }
        for (const_iterator curr = begin(); curr != end(); ++curr) {
            if (!h.locate((*curr).first).first) {
                return true;
            }
        }
        return false;
    }


    local_hash_table(key_hash hf) :
        hashf_(hf), chain_size_(DEFAULT_CHAIN_SIZE),
        data_(DEFAULT_NBUCKETS), llevel_(DEFAULT_NBUCKETS, 0),
        nbuckets_(DEFAULT_NBUCKETS), glevel_(0), next_(0) {
    }

    local_hash_table(const_iterator from, const_iterator last, key_hash hf) :
        hashf_(hf), chain_size_(DEFAULT_CHAIN_SIZE),
        data_(DEFAULT_NBUCKETS), llevel_(DEFAULT_NBUCKETS, 0),
        nbuckets_(DEFAULT_NBUCKETS), glevel_(0), next_(0) {
        for (; from != last; ++from) {
            insert(*from);
        }
    }

    local_hash_table(const self& h) :
        hashf_(h.hashf_), chain_size_(h.chain_size_),
        data_(h.data_), llevel_(h.llevel_),
        nbuckets_(h.nbuckets_), glevel_(h.glevel_), next_(h.next_) {
    }

    const self& operator=(const self& h) {
        if (this != &h) {
            hashf_      = h.hashf_;
            chain_size_ = h.chain_size_;

            data_       = h.data_;
            llevel_     = h.llevel_;

            nbuckets_   = h.nbuckets_;
            glevel_     = h.glevel_;
            next_       = h.next_;
        }
        return *this;
    }

    ~local_hash_table() {
    }
};





#endif /* !defined(_hash_table_h_) */
