/************************************************************************
 * ss_symtab.cc: symbol table manipulation for Simple-SUIF.
************************************************************************/





/************************************************************************
 * header files.
************************************************************************/

#include <iomanip.h>
#include <iostream.h>
#include <string>
#include <suif.h>

extern "C" {
#include <simple.h>
#include <stdio.h>
}

#include "hash_map.h"
#include "ss_symtab.h"





/************************************************************************
 * types.
************************************************************************/

typedef hash_map<string, ss_symbol, hash<string> > ss_symtab_t;





/************************************************************************
 * struct ss_proc_symtab
************************************************************************/

struct ss_proc_symtab {
    string      proc_name;
    ss_symtab_t proc_symtab;
    unsigned    nproc_symbols;
    ss_symbol*  proc_symbols;
    unsigned    nparams;
    ss_symbol*  params;

    ss_proc_symtab(const string& n) :
        proc_name(n), proc_symtab(STRINGS::string_hash), proc_symbols(0),
        nparams(0), params(0) {
    }

};

typedef hash_map<string, ss_proc_symtab *, hash<string> > ss_proc_symtab_t;





/************************************************************************
 * variables.
************************************************************************/

ss_symbol* ss_global_syms  = 0;
unsigned   ss_global_nsyms = 0;

static ss_symtab_t      ss_globals(STRINGS::string_hash);
static ss_proc_symtab_t ss_proc_symbols(STRINGS::string_hash);





/************************************************************************
 * void generate_data_definition(const ss_symbol& sym)
************************************************************************/

static
void
generate_data_definition(const ss_symbol& sym, var_def* var) {
    cout << "!"                          << endl
         << "! data segment definitions" << endl
         << "!"                          << endl << endl
         << "\t.data"                    << endl << endl;

    if (var->variable()->is_global()) {
        cout << "\t.global\t"
#if defined(PREFIX_UNDERSCORE)
        << "_"
#endif /* defined(PREFIX_UNDERSCORE) */
        << sym.sym_name << endl;
    }
    cout << "\t.align\t"        << ((sym.sym_align == 0) ? 4 : sym.sym_align/8)
                                       << endl;
    cout
#if defined(PREFIX_UNDERSCORE)
        << "_"
#endif /* defined(PREFIX_UNDERSCORE) */
        << sym.sym_name << ":" << endl;

    annote_list_iter anli(var->annotes());
    if (anli.is_empty()) {
        cout << "\t.skip\t" << sym.sym_size/8 << endl;
    }
    else {
        while (!anli.is_empty()) {
            annote* ann = anli.step();
            if (ann->name() == k_fill) {
                immed_list*     iml = ann->immeds();
                immed_list_iter imli(iml);
                unsigned        sz = imli.step().unsigned_int()/8;
                cout << "\t.skip\t" << sz << endl;
            }
            else if (ann->name() == k_repeat_init) {
                immed_list*     iml = ann->immeds();
                immed_list_iter imli(iml);
                unsigned count = imli.step().unsigned_int();
                unsigned sz  = imli.step().unsigned_int()/8;
                immed    imm = imli.step();
                for (unsigned i = 0; i < count; ++i) {
                    switch (imm.kind()) {
                        case im_int: {
                            int val = imm.integer();
                            cout << "\t." << ((sz == 1) ? "byte"
                                : ((sz == 2) ? "half"
                                : ((sz == 4) ? "word" : "BOGUS")))
                                << "\t" << val << endl;
                            break;
                        }

                        case im_float: {
                            if (sz == 4) {
                                float val = imm.flt();
                                int*  iv  = (int *) (&val);
                                cout << "\t.word\t0x" << hex << (*iv) << dec << endl; ++iv;
                            }
                            else if (sz == 8) {
                                double val = imm.flt();
                                int*   iv  = (int *) (&val);
                                cout << "\t.word\t0x" << hex << (*iv) << dec << endl; ++iv;
                                cout << "\t.word\t0x" << hex << (*iv) << dec << endl; ++iv;
                            }
                            else {
                                cerr << "*** cannot handle fp size " << sz
                                     << " in repeat_init ***" << endl;
                            }
                            break;
                        }

                        case im_symbol: {
                            sym_node* sym = imm.symbol();
                            int off = int(imm.addr().offset())/8;
                            if (off != int(imm.offset()/8)) {
                                cerr << "*** mismatched offsets ***" << endl;
                            }
                            cout << "\t.word\t"
#if defined(PREFIX_UNDERSCORE)
                                 << "_"
#endif /* defined(PREFIX_UNDERSCORE) */
                                 << sym->name();
                            if (off > 0) {
                                cout << "+" << off;
                            }
                            else if (off < 0) {
                                cout << off;
                            }
                            cout << endl;

                            break;
                        }

                        case im_extended_int: {
                            if (sz != 4) {
                                cerr << "*** why is extended int not a word ***" << endl;
                            }
                            else {
                                cout << "\t.word\t" << imm.ext_integer() << endl;
                            }
                            break;
                        }

                        default:
                            cerr << "*** cannot generate definition for repeat_init ***"
                                 << endl;
                            break;
                    }
                }
            }
            else if (ann->name() == k_multi_init) {
                immed_list*     iml = ann->immeds();
                immed_list_iter imli(iml);
                unsigned        sz = imli.step().unsigned_int()/8;
                if (sz != 1) {
                    cerr << "*** multi_init of size " << sz << " ***" << endl;
                }
                else {
                    cout << "\t.ascii\t\"";
                    while (!imli.is_empty()) {
                        char ch = imli.step().integer();
                        if ((ch == '"') || (ch == '\\')) {
                            cout << "\\" << ch;
                        }
                        else if (ch == '\0') {
                            cout << "\\0";
                        }
                        else if (ch == '\n') {
                            cout << "\\n";
                        }
                        else {
                            cout << ch;
                        }
                    }
                    cout << "\"" << endl;
                }
            }
            else {
                cerr << "*** cannot handle annotation " << ann->name() << " ***" << endl;
            }
        }
    }

    cout << endl;
}





/************************************************************************
 * void process_symbols_and_definitions(sym_node_list* syms,
 *                                      var_def_list* vars,
 *                                      ss_symtab_t& tab)
************************************************************************/

static
void
process_symbols_and_definitions(sym_node_list* syms, var_def_list* vars,
    ss_symtab_t& tab) {
    if (!syms->is_empty()) {
        sym_node_list_iter si(syms);
        while (!si.is_empty()) {
            sym_node* sym = si.step();
            if (tab.find(sym->name()) != tab.end()) {
                cerr << "*** symbol " << sym->name() << " already processed!!! ***"
                     << endl;
            }
            else {
                tab[sym->name()].sym_name     = sym->name();
                tab[sym->name()].sym_size     = ((sym->kind() == SYM_VAR)
                                                  ? ((var_sym *)sym)->type()->size()
                                                  : 0);
                tab[sym->name()].sym_align    = 0;
                tab[sym->name()].is_defined   = 0;
                tab[sym->name()].is_aggregate =
                    ((sym->kind() == SYM_VAR)
                        ? (((var_sym *)sym)->type()->is_struct() == true)
                        : 0);
            }
        }
    }

    if (!vars->is_empty()) {
        var_def_list_iter vi(vars);
        while (!vi.is_empty()) {
            var_def* var = vi.step();
            var_sym* sym = var->variable();
            if (tab.find(sym->name()) == tab.end()) {
                cerr << "*** variable " << sym->name() << " not declared!!! ***"
                     << endl;
            }
            else {
                tab[sym->name()].sym_size     = sym->type()->size();
                tab[sym->name()].sym_align    = var->alignment();
                tab[sym->name()].is_defined   = 1;
                tab[sym->name()].is_aggregate = sym->type()->is_struct();
                generate_data_definition(tab[sym->name()], var);
            }
        }
    }
}





/************************************************************************
 * functions callable from Simple-SUIF
************************************************************************/

static
void
make_symbol_array(const ss_symtab_t& tab, unsigned& size, ss_symbol*& syms) {
    size  = tab.size();
    syms  = new ss_symbol[size];
    unsigned i = 0;
    ss_symtab_t::const_iterator ti = tab.begin();
    for (; (i < size) && (ti != tab.end()); ++i, ++ti) {
        syms[i].sym_name     = (*ti).second.sym_name;
        syms[i].sym_size     = (*ti).second.sym_size;
        syms[i].sym_align    = (*ti).second.sym_align;
        syms[i].is_defined   = (*ti).second.is_defined;
        syms[i].is_aggregate = (*ti).second.is_aggregate;
    }
}

extern "C" {
void
ss_globals_init(sym_node_list* syms, var_def_list* vars) {
    process_symbols_and_definitions(syms, vars, ss_globals);
}

void
ss_file_init(sym_node_list* syms, var_def_list* vars) {
    process_symbols_and_definitions(syms, vars, ss_globals);
    make_symbol_array(ss_globals, ss_global_nsyms, ss_global_syms);
}

void
ss_proc_init(const char* proc, sym_node_list* syms,
    var_def_list* vars, sym_node_list* parms) {
    if (ss_proc_symbols.find(proc) != ss_proc_symbols.end()) {
        cerr << "*** procedure " << proc
             << " symbol table already defined!!! ***" << endl;
    }
    else {
        ss_proc_symbols[proc] = new ss_proc_symtab(proc);
    }

    process_symbols_and_definitions(syms, vars,
        ss_proc_symbols[proc]->proc_symtab);
    make_symbol_array(ss_proc_symbols[proc]->proc_symtab,
        ss_proc_symbols[proc]->nproc_symbols,
        ss_proc_symbols[proc]->proc_symbols);

    unsigned&   nparams = ss_proc_symbols[proc]->nparams;
    ss_symbol*& params  = ss_proc_symbols[proc]->params;
    sym_node_list_iter si(parms);
    while (!si.is_empty()) {
        ++nparams; si.step();
    }

    si.reset(parms);
    params = new ss_symbol[nparams];
    unsigned i = 0;
    while (!si.is_empty()) {
        sym_node* sym = si.step();

        params[i].sym_name     = sym->name();
        params[i].sym_size     = ((sym->kind() == SYM_VAR)
                                ? (((var_sym *)sym)->type()->size())
                                : (0));
        params[i].sym_align    = 0;
        params[i].is_defined   = 0;
        params[i].is_aggregate = ((sym->kind() == SYM_VAR)
                                ? (((var_sym *)sym)->type()->is_struct() == true)
                                : (0));

        ++i;
    }
}
};





/************************************************************************
 * functions callable from usercode.
************************************************************************/

extern "C" {
const ss_symbol*
ss_proc_syms(const char* proc) {
    if (ss_proc_symbols.find(proc) == ss_proc_symbols.end()) {
        cerr << "*** no symbol table available for procedure " << proc
             << "***" << endl;
        return 0;
    }
    return ss_proc_symbols[proc]->proc_symbols;
}

unsigned
ss_proc_nsyms(const char* proc) {
    if (ss_proc_symbols.find(proc) == ss_proc_symbols.end()) {
        cerr << "*** no symbol table available for procedure " << proc
             << "***" << endl;
        return 0;
    }
    return ss_proc_symbols[proc]->nproc_symbols;
}

const ss_symbol*
ss_proc_params(const char* proc) {
    if (ss_proc_symbols.find(proc) == ss_proc_symbols.end()) {
        cerr << "*** no parameter list available for procedure " << proc
             << "***" << endl;
        return 0;
    }
    return ss_proc_symbols[proc]->params;
}

unsigned
ss_proc_nparams(const char* proc) {
    if (ss_proc_symbols.find(proc) == ss_proc_symbols.end()) {
        cerr << "*** no parameter list available for procedure " << proc
             << "***" << endl;
        return 0;
    }
    return ss_proc_symbols[proc]->nparams;
}
};





/************************************************************************
 * unsigned get_referrent_size(const simple_reg* sreg)
************************************************************************/

extern "C" {
unsigned
get_referrent_size(const simple_reg* sreg) {
    type_node* ty = (type_node *) (sreg->var->type->suif_type);
    if (!ty->is_ptr()) {
        cerr << "*** SUIF type is not a pointer ***" << endl;
        return 0;
    }

    ptr_type* pt = (ptr_type *) (sreg->var->type->suif_type);
    return pt->ref_type()->size();
}
};
